package org.artifactory.storage.service;

import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.HaAddon;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.storage.BinariesInfo;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.mbean.MBeanRegistrationService;
import org.artifactory.model.common.RepoPathImpl;
import org.artifactory.model.xstream.fs.FileInfoImpl;
import org.artifactory.projects.ProjectsService;
import org.artifactory.schedule.TaskService;
import org.artifactory.spring.InternalArtifactoryContext;
import org.artifactory.storage.StorageSummaryInfo;
import org.artifactory.storage.StorageUsage;
import org.artifactory.storage.binstore.service.InternalBinaryService;
import org.artifactory.storage.db.DbService;
import org.artifactory.storage.db.fs.model.DbMutableFile;
import org.artifactory.storage.fs.repo.CacheUnAvailableException;
import org.artifactory.storage.fs.repo.StoringRepo;
import org.artifactory.storage.fs.service.ConfigsService;
import org.artifactory.storage.fs.service.FileService;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.jfrog.storage.DbType;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.Assert;
import org.testng.annotations.AfterClass;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicLong;

import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

/**
 * @author barh
 */
public class StorageServiceImplTest extends ArtifactoryHomeBoundTest {

    private final static String PROJECT = "projectx";
    public static final String PROJ_REPO = "source-repo";

    @Mock
    private CentralConfigService centralConfigService;
    @Mock
    private DbService dbService;
    @Mock
    private InternalBinaryService binaryStore;
    @Mock
    private FileService fileService;
    @Mock
    private TaskService taskService;
    @Mock
    private RepositoryService repositoryService;
    @Mock
    private AddonsManager addonsManager;
    @Mock
    private ProjectsService projectsService;
    @Mock
    private ConfigsService configsService;
    @Mock
    InternalArtifactoryContext context;

    @BeforeClass
    public void init() {
        bindArtifactoryHome();
        MockitoAnnotations.initMocks(this);
        MBeanRegistrationService mbeanService = mock(MBeanRegistrationService.class);
        when(context.beanForType(MBeanRegistrationService.class)).thenReturn(mbeanService);
        ArtifactoryContextThreadBinder.bind(context);
    }

    @AfterClass
    public void cleanup() {
        ArtifactoryContextThreadBinder.unbind();
        unbindArtifactoryHome();
    }

    @Test
    void testGetCachedStorageSizeResolveFromDB() throws SQLException {
        //prepare
        StorageServiceImpl storageHalfMock = spy(StorageServiceImpl.class);
        doThrow(CacheUnAvailableException.class).when(storageHalfMock).getStorageSummaryInfoFromCache();
        InternalBinaryService internalBinaryService = mock(InternalBinaryService.class);
        when(internalBinaryService.getBinariesInfo()).thenReturn(mock(BinariesInfo.class));
        storageHalfMock.setInternalBinaryService(internalBinaryService);

        //act
        when(storageHalfMock.getCachedStorageSize()).thenCallRealMethod();

        //verify
        verify(internalBinaryService, times(1)).getBinariesInfo();
    }

    @Test
    void testGetCachedStorageSizeResolveFromCache() throws SQLException {
        //prepare
        StorageServiceImpl storageHalfMock = spy(StorageServiceImpl.class);
        StorageSummaryInfo storageSummaryInfo = mock(StorageSummaryInfo.class);
        when(storageSummaryInfo.getBinariesInfo()).thenReturn(mock(BinariesInfo.class));
        doReturn(storageSummaryInfo).when(storageHalfMock).getStorageSummaryInfoFromCache();
        InternalBinaryService internalBinaryService = mock(InternalBinaryService.class);
        storageHalfMock.setInternalBinaryService(internalBinaryService);

        //act
        when(storageHalfMock.getCachedStorageSize()).thenCallRealMethod();

        //verify
        verify(internalBinaryService, times(0)).getBinariesInfo();
    }

    @Test
    void testGetProjectStorageSum() {
        StorageServiceImpl storageService = initStorageService();
        Assert.assertEquals(1000L, storageService.getProjectStorageUsage(PROJECT));
        Assert.assertEquals(0L, storageService.getProjectStorageUsage(ProjectsService.DEFAULT_PROJECT));
    }

    @Test
    void testSyncStorageUsageDiffsIfNeeded() {
        // Setup
        StorageServiceImpl storageService = initStorageService();
        when(projectsService.getAllProjects()).thenReturn(List.of(PROJECT));
        when(projectsService.getProjectOfRepository(PROJ_REPO)).thenReturn(PROJECT);
        doThrow(IllegalArgumentException.class).when(storageService).saveAndReloadStorageUsageDiffsWithRetries();
        HaAddon haAddon = mock(HaAddon.class);
        when(haAddon.getAllArtifactoryServers()).thenReturn(new ArrayList<>());
        when(addonsManager.addonByType(HaAddon.class)).thenReturn(haAddon);

        // Set the project limit for 2000 bytes
        when(projectsService.getProjectStorageQuotaBytes(PROJECT)).thenReturn(2000L);
        // Call syncStorageUsageDiffsIfNeeded and expect nothing to happened
        storageService.syncStorageUsageDiffsIfNeeded();

        //Deploy a small file into a project related repository
        storageService.afterCreate(prepareDbMutableFileMock(1), null, null);
        // Call syncStorageUsageDiffsIfNeeded and expect nothing to happened
        storageService.syncStorageUsageDiffsIfNeeded();

        // Deploy a bigger file into the project related repository
        storageService.afterCreate(prepareDbMutableFileMock(500), null, null);
        try {
            // Call syncStorageUsageDiffsIfNeeded and expect saveAndReloadStorageUsageDiffsWithRetries will get called
            storageService.syncStorageUsageDiffsIfNeeded();
            Assert.fail("saveAndReloadStorageUsageDiffsWithRetries was never called");
        } catch (IllegalArgumentException ignored) {}

        // Delete the file from before so the total diff will be only 1 byte
        storageService.afterDelete(prepareDbMutableFileMock(500), null, null);
        // Call syncStorageUsageDiffsIfNeeded for 15 times in total (including the previous retries) and expect
        // saveAndReloadStorageUsageDiffsWithRetries will get called at the last try
        for (int i = 0; i < 11; i++) {
            storageService.syncStorageUsageDiffsIfNeeded();
        }
        try {
            storageService.syncStorageUsageDiffsIfNeeded();
            Assert.fail("saveAndReloadStorageUsageDiffsWithRetries was never called");
        } catch (IllegalArgumentException ignored) {}
    }

    @Test
    void testUpdateStorageSummaryWithStorageUsage() {
        StorageServiceImpl storageService = initStorageService();
        StorageUsage storageUsage = new StorageUsage();
        storageUsage.getProjectsStorageSum().put("project", 1000L);
        storageUsage.getRepoStorageSum().put("proj-repo", 1000L);
        storageService.updateStorageSummaryWithStorageUsage(storageUsage);

        StorageSummaryInfo cache = storageService.getStorageSummaryInfoFromCache();
        Assert.assertEquals(1000L, cache.getProjectsTotalSizes().getOrDefault("project", 0L).longValue());
        Assert.assertEquals(1000L, cache.getRepoStorageSummaries().get("proj-repo").getUsedSpace());
    }

    @Test
    void testMergeStorageDiffWithStorageUsage() {
        StorageServiceImpl storageService = initStorageService();
        when(projectsService.getProjectOfRepository(PROJ_REPO)).thenReturn(PROJECT);

        StorageUsage oldStorageUsage = new StorageUsage();
        oldStorageUsage.setRevision(1);
        oldStorageUsage.getProjectsStorageSum().put(PROJECT, 1000L);
        oldStorageUsage.getRepoStorageSum().put("other-proj-repo", 1000L);

        // Deploy some file so we gonna have some changes to merge
        storageService.afterCreate(prepareDbMutableFileMock(500), null, null);

        // Call mergeStorageDiffWithStorageUsage
        StorageUsage mergedStorageUsage = storageService.mergeStorageDiffWithStorageUsage(oldStorageUsage);

        // Verify the merged StorageUsage contains both local changes and oldStorageUsage
        Assert.assertEquals(mergedStorageUsage.getProjectsStorageSum().get(PROJECT).longValue(), 1500);
        Assert.assertEquals(mergedStorageUsage.getRepoStorageSum().get(PROJ_REPO).longValue(), 500);
        Assert.assertEquals(mergedStorageUsage.getRepoStorageSum().get("other-proj-repo").longValue(), 1000);
        Assert.assertEquals(mergedStorageUsage.getRevision(), 2);
    }

    @Test
    void testSyncStorageUsageWithStorageSummary() {
        StorageServiceImpl storageService = initStorageService();
        when(configsService.getConfig(ArtifactoryHome.STORAGE_USAGE_FILE)).thenReturn("{\"revision\":1603365765723," +
                "\"repoStorageSum\":{\"example-repo-local\":5703478,\"prjx-generic\":43741906," +
                "\"artifactory-build-info\":0,\"auto-trashcan\":0,\"jfrog-support-bundle\":0}," +
                "\"projectsStorageSum\":{\"default\":5703478,\"projectx\":43741906}}");
        // Sync
        storageService.syncStorageUsageWithStorageSummary();

        // Verify the local cache has been updated correctly
        StorageSummaryInfo cache = storageService.getStorageSummaryInfoFromCache();
        Assert.assertEquals(43741906, cache.getProjectsTotalSizes().get(PROJECT).longValue());
        Assert.assertEquals(43741906, cache.getRepoStorageSummaries().get("prjx-generic").getUsedSpace());
        Assert.assertEquals(5703478, cache.getProjectsTotalSizes().get(ProjectsService.DEFAULT_PROJECT).longValue());
        Assert.assertEquals(5703478, cache.getRepoStorageSummaries().get("example-repo-local").getUsedSpace());
        Assert.assertEquals(0, cache.getRepoStorageSummaries().get("auto-trashcan").getUsedSpace());
    }

    @Test
    void testMergeDiffChanges() {
        StorageServiceImpl storageService = initStorageService();
        Map<String, AtomicLong> realDiffMap = new ConcurrentHashMap<>();
        Map<String, AtomicLong> backupDiffMap = new HashMap<>();
        // Create some data to merge
        realDiffMap.put("repo", new AtomicLong(10000L));
        backupDiffMap.put("repo", new AtomicLong(99L));
        backupDiffMap.put("repo2", new AtomicLong(100L));
        // Act
        storageService.mergeDiffChanges(backupDiffMap, realDiffMap);
        // Validate the merged data contains both source and target maps
        Assert.assertEquals(realDiffMap.get("repo").get(), 10099L);
        Assert.assertEquals(realDiffMap.get("repo2").get(), 100L);
    }

    private StorageServiceImpl initStorageService() {
        StorageServiceImpl storageService = spy(new StorageServiceImpl());
        storageService.setAddonsManager(addonsManager);
        storageService.setBinaryStore(binaryStore);
        storageService.setCentralConfigService(centralConfigService);
        storageService.setDbService(dbService);
        storageService.setFileService(fileService);
        storageService.setTaskService(taskService);
        storageService.setRepositoryService(repositoryService);
        storageService.setProjectsService(projectsService);
        storageService.setConfigsService(configsService);
        when(dbService.getDatabaseType()).thenReturn(DbType.MYSQL);
        CentralConfigDescriptor centralConfigDescriptor = mock(CentralConfigDescriptor.class);
        when(centralConfigService.getDescriptor()).thenReturn(centralConfigDescriptor);
        doNothing().when(storageService).rescheduleGC(centralConfigDescriptor, null);
        doReturn(true).when(storageService).isProjectsStorageQuotaEnabled();
        storageService.init();
        populateStorageSummaryCache();
        storageService.updateStorageSummaryCache(populateStorageSummaryCache());
        return storageService;
    }

    private StorageSummaryInfo populateStorageSummaryCache() {
        StorageSummaryInfo summaryInfo = new StorageSummaryInfo(new HashMap<>(), null, new HashMap<>());
        summaryInfo.getProjectsTotalSizes().putIfAbsent(PROJECT, 1000L);
        return summaryInfo;
    }

    public static DbMutableFile prepareDbMutableFileMock(long size) {
        RepoPathImpl repoPath = new RepoPathImpl(PROJ_REPO, "test.file");
        FileInfoImpl fileInfo = new FileInfoImpl(repoPath);
        fileInfo.setSize(size);
        return new DbMutableFile(mock(StoringRepo.class), 52437842367L, fileInfo);
    }
}