/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2019 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.logging.converter;

import lombok.AccessLevel;
import lombok.NoArgsConstructor;
import org.artifactory.util.StringInputStream;
import org.artifactory.util.XmlUtils;
import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.JDOMException;
import org.jdom2.Namespace;
import org.jdom2.Text;
import org.jdom2.input.SAXBuilder;
import org.jfrog.common.BiOptional;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import java.util.function.Consumer;
import java.util.function.Predicate;

/**
 * @author Shay Bagants
 * @author Dan Feldman
 */
@NoArgsConstructor(access = AccessLevel.PRIVATE)
public class LogbackConverterUtils {
    private static final Logger log = LoggerFactory.getLogger(LogbackConverterUtils.class);

    public static final String ELEMENT_TYPE_APPENDER = "appender";
    public static final String ELEMENT_TYPE_LOGGER = "logger";
    public static final String ATTRIBUTE_NAME = "name";

    /**
     * Appends {@param appenderContent} with {@param appenderName} to the end of the appenders list in the xml
     */
    public static void addAppender(Element root, Namespace ns, String appenderName, String appenderContent) {
        addElement(root, ns, ELEMENT_TYPE_APPENDER, appenderName, appenderContent);
    }

    /**
     * Appends {@param loggerContent} with {@param loggerName} to the end of the loggers list in the xml
     */
    public static void addLogger(Element root, Namespace ns, String loggerName, String loggerContent) {
        addElement(root, ns, ELEMENT_TYPE_LOGGER, loggerName, loggerContent);
    }

    /**
     * Appends {@param elementContent} with {@param elementName} to the end of the loggers list in the xml
     */
    private static void addElement(Element root, Namespace ns, String elementType, String elementName,
            String elementContent) {
        List<Element> elements = root.getChildren(elementType, ns);
        performIfElementNotPresentElement(root, ns, elementType,
                elementNameEquals(ns, elementName),
                element -> log
                        .info("{} {} already exists in logback.xml, skipping conversion", elementType, elementName),
                () -> appendElementContent(root, elementName, elementContent, elements)
        );
    }

    private static void appendElementContent(Element root, String elementName, String elementContent,
            List<Element> currentElements) {
        try {
            Element logger = getElement(elementContent);
            root.addContent(root.indexOf(currentElements.get(currentElements.size() - 1)) + 1, new Text("\n    "));
            root.addContent(root.indexOf(currentElements.get(currentElements.size() - 1)) + 2, logger);
        } catch (Exception e) {
            logError(e, elementName);
        }
    }

    /**
     * Removes {@param appenderName} if it exists in the list of appenders in this {@param root} element.
     */
    public static void removeAppenderIfExists(Element root, Namespace ns, String appenderName) {
        performForElement(root, ns, ELEMENT_TYPE_APPENDER,
                elementNameEquals(ns, appenderName),
                element -> {
                    log.info("removing old appender: {}", appenderName);
                    root.removeContent(element);
                }
        );
    }

    /**
     * Removes {@param loggerName} if it exists in the list of loggers in this {@param root} element.
     */
    public static void removeLoggerIfExists(Element root, Namespace ns, String loggerName) {
        performForElement(root, ns, ELEMENT_TYPE_LOGGER,
                elementNameEquals(ns, loggerName),
                element -> {
                    log.info("removing old logger: {}", loggerName);
                    root.removeContent(element);
                }
        );
    }

    /**
     * Return Element object from String
     */
    public static Element getElement(String input) throws IOException, JDOMException {
        SAXBuilder builder = XmlUtils.createSaxBuilder();
        try (InputStream stream = new StringInputStream(input)) {
            Document doc = builder.build(stream);
            Element root = doc.getRootElement();
            return root.detach();
        }
    }

    private static void performForElement(Element root, Namespace ns, String elementType, Predicate<Element> filter,
            Consumer<Element> action) {
        root.getChildren(elementType, ns).stream()
                .filter(filter)
                .findAny()
                .ifPresent(action);
    }

    private static void performIfElementNotPresentElement(Element root, Namespace ns, String elementType,
            Predicate<Element> filter, Consumer<Element> elementExistsAction, Runnable action) {
        BiOptional.of(root.getChildren(elementType, ns).stream()
                .filter(filter)
                .peek(elementExistsAction)
                .findAny())
                .ifNotPresent(action);
    }


    private static Predicate<Element> elementNameEquals(Namespace ns, String elementName) {
        return element -> element.getAttributeValue(ATTRIBUTE_NAME, ns).equals(elementName);
    }

    public static void logError(Exception e, String elementName) {
        String err = "Error processing '" + elementName + "' element in logback.xml: ";
        log.error("{} {}", err, e.getMessage());
        log.error("Your logback configuration is not in the desired state, please reference the default example " +
                "supplied with the application and manually fix accordingly.");
        log.debug(err, e);
    }
}
