#!/usr/bin/env bash

set -e

export PROJECT_ROOT_DIR="$(cd "$(dirname "${BASH_SOURCE[0]}")" && cd ../../.. && pwd)"
source "${PROJECT_ROOT_DIR}/build/ci/scripts/lib/commonUtils.sh"
source "${PROJECT_ROOT_DIR}/build/ci/scripts/lib/mavenUtils.sh"
source "${PROJECT_ROOT_DIR}/build/ci/scripts/lib/dockerUtils.sh"
source "${PROJECT_ROOT_DIR}/build/ci/scripts/lib/versionUtils.sh"
source "${PROJECT_ROOT_DIR}/build/ci/scripts/lib/pipelineUtils.sh"
source "${PROJECT_ROOT_DIR}/build/ci/scripts/lib/buildInfoUtils.sh"


function on_start() {
  cd "${PROJECT_ROOT_DIR}/artifactory-oss"
  setCustomBuildTypes
  buildInfo_setBuildNameAndNumber
  ${PROJECT_ROOT_DIR}/build/ci/scripts/javaGenerateConfig.sh
  if ! command -v "rpmbuild"; then
    echo "rpmbuild tool is missing - installing using apt-get..."
    apt-get update && apt-get install -y rpm rsync
    if ! command -v "rpmbuild" ; then
      echo "rpmbuild is not available in path, exiting..."
      exit 1
    fi
  fi

  echo "Calculating TARGET_VERSION and DOCKER_COMPOSE_REGISTRY for the docker compose env"
  cd "${PROJECT_ROOT_DIR}"
  local version targetVersion
  version="$(maven_getVersionFromPom "artifactory-oss/pom.xml")"
  echo "Version: ${version}"
  targetVersion="${version}"
  if version_isSnapshot "${version}"; then
    local branch normalizedBranch
    branch="$(pipeline_gitBranchName)"
    echo "Branch:  ${branch}"
    normalizedBranch=$(common_normalizeBranchName "${branch}")
    targetVersion="${version}-${normalizedBranch}-${run_number}"
  fi
  local dockerComposeRegistry="docker.jfrog.io"
  if version_isRelease "${version}"; then
    dockerComposeRegistry="docker.bintray.io"
  fi
  echo "Setting following env for the docker compose:"
  echo "  TARGET_VERSION=${targetVersion}"
  echo "  DOCKER_COMPOSE_REGISTRY=${dockerComposeRegistry}"
  add_run_variables TARGET_VERSION="${targetVersion}" DOCKER_COMPOSE_REGISTRY="$dockerComposeRegistry"
  cd "${PROJECT_ROOT_DIR}/artifactory-oss"
}

function on_execute() {
  return 0
}

function on_complete() {
  return 0
}

function on_success() {
  # Publish the build with the Maven artifacts
  local version
  version=$(maven_getVersionFromPom "pom.xml")
  buildInfo_modifyBuildName "${JFROG_CLI_BUILD_NAME}" "${JFROG_CLI_BUILD_NUMBER}"
  buildInfo_publish "${JFROG_CLI_BUILD_NAME}"
  write_output artifactory_oss_package_build_info buildName="${JFROG_CLI_BUILD_NAME}" buildNumber="${JFROG_CLI_BUILD_NUMBER}" version="${version}"

  # Push and publish a build with the docker images
  local buildName
  buildName="$(docker_generateBuildName)"
  local repository="art-docker-dev-local"
  local imageName="jfrog/artifactory-oss"
  local fullImageName imageTag
  fullImageName="$(docker_tag "${imageName}" "art_jfrog_io")"
  docker_push "${fullImageName}" "${buildName}" "${JFROG_CLI_BUILD_NUMBER}" "${repository}"
  imageTag=$(docker_extractImageTag "$fullImageName")
  write_output artifactory_oss_art_jfrog_io_docker_image sourceRepository="${repository}" imageName="${imageName}" imageTag="${imageTag}"

  buildInfo_publish "${buildName}"
  write_output artifactory_oss_docker_build_info buildName="${buildName}" buildNumber="${JFROG_CLI_BUILD_NUMBER}"
}

# exports an environment variable with desired build types based on the last git commit body or env variable
function setCustomBuildTypes(){
  local branch
  branch="$(pipeline_gitBranchName)"
  if [[ $(common_gitLastCommitBody) == *"[package-all-types]"* ]] || [[ "${PACKAGE_ALL_TYPES}" == true ]] || [[ "$branch" =~ ^release\/|^preRelease\/|master ]]; then
    add_run_variables ARTIFACTORY_OSS_BUILD_TYPES="release,docker,deb,rpm"
  else
    add_run_variables ARTIFACTORY_OSS_BUILD_TYPES="release,docker"
  fi
  export_run_variables
}